--[[-------------------------------------------------------------------------
	QuickTargetMacros
    Copyright (C) 2010  Morsker
	Please contact Morsker through PM on forums.wowace.com.

	Aids the user in quickly making /target macros for immediate use.

	License:
		This program is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License as published by
		the Free Software Foundation, either version 3 of the License, or
		(at your option) any later version.

		This program is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with this program. If not, see <http://www.gnu.org/licenses/>.

	Note:
		This AddOn's source code is specifically designed to work with
		World of Warcraft's interpreted AddOn system.
		You have an implicit license to use this AddOn with these facilities
		since that is its designated purpose as per:
		http://www.fsf.org/licensing/licenses/gpl-faq.html#InterpreterIncompat
---------------------------------------------------------------------------]]
QTMacros = LibStub("AceAddon-3.0"):NewAddon("QTMacros", "AceConsole-3.0", "AceEvent-3.0")
local QTMacros = QTMacros
local L = LibStub("AceLocale-3.0"):GetLocale("QTMacros", true)
local db
local mainButton
local keybindFrame
local ShowKeybindFrame -- local function, forward declaration

local QTM_DB_VERSION = 1 -- This is not the release version. It's for db versioning.


function BuildMacro(template, targets)
	local t = {}
	tinsert(t, template.header and template.header or "")
	local start = 1
	if template.firstLine and #targets >= 1 then
		tinsert(t, (gsub(template.firstLine, "$s", targets[1])))
		tinsert(t, "\n")
		start = 2
	end
	if template.line then
		for i = start,#targets do
			tinsert(t, (gsub(template.line, "$s", targets[i])))
			tinsert(t, "\n")
		end
	end
	tinsert(t, template.footer and template.footer or "")
	return table.concat(t)
end


--------------------
-- AceConfig Options
--------------------

local options = nil
local function giveOptions()
	if options then return options end
	options = {
		name = "QTMacros",
		handler = QTMacros,
		type = "group",
		args = {
			bind = {
				name = "Bind",
				order = 0,
				type = "execute",
				dialogHidden = true,
				desc = "Choose a keybind for the macro.",
				func = function(info) ShowKeybindFrame() end,
			},
			show = {
				name = "Show",
				order = 1,
				type = "execute",
				dialogHidden = true,
				desc = "Show the QT button.",
				func = function(info) QTMacros:ActivateButton() end,
			},
			config = {
				name = "Config",
				order = 2,
				type = "execute",
				dialogHidden = true,
				desc = "Open config UI.",
				func = function(info) InterfaceOptionsFrame_OpenToCategory("QuickTargetMacros") end,
			},
			header_header = {
				name = "Header",
				order = 10,
				type = "header",
			},
			header = {
				name = "Header",
				order = 11,
				type = "input",
				multiline = true,
				cmdHidden = true,
				desc = "macro header, before targetting",
				set = function(info, val)
					if #val == 0 then
						val = nil
					elseif not strmatch(val, "\n$") then
						val = val.."\n"
					end
					db.cmdTemplate.header = val
				end,
				get = function(info) return db.cmdTemplate.header end,
			},
			firstLine_header = {
				name = "First Line",
				order = 12,
				type = "header",
			},
			firstLine = {
				name = "First Line",
				order = 13,
				type = "input",
				cmdHidden = true,
				desc = "use this if the first line should be different; leave blank otherwise",
				set = function(info, val) db.cmdTemplate.firstLine = (#val > 0 and val or nil) end,
				get = function(info) return db.cmdTemplate.firstLine end,
			},
			line_header = {
				name = "Line",
				order = 14,
				type = "header",
			},
			line = {
				name = "Line",
				order = 15,
				type = "input",
				cmdHidden = true,
				desc = "template for each line of targetting,  use $s to include the name",
				set = function(info, val) db.cmdTemplate.line = (#val > 0 and val or nil) end,
				get = function(info) return db.cmdTemplate.line end,
			},
			footer_header = {
				name = "Footer",
				order = 16,
				type = "header",
			},
			footer = {
				name = "Footer",
				order = 17,
				type = "input",
				multiline = true,
				cmdHidden = true,
				desc = "macro footer, after targetting",
				set = function(info, val)
					if #val == 0 then
						val = nil
					elseif not strmatch(val, "\n$") then
						val = val.."\n"
					end
					db.cmdTemplate.footer = val
				end,
				get = function(info) return db.cmdTemplate.footer end,
			},
		},
	}
	return options
end


-----------------
-- addon QTMacros
-----------------


function QTMacros:GetLineTemplate(info)
	return lineTemplate
end


function QTMacros:SetLineTemplate(info, input)
	lineTemplate = input
end


-- These generate OnKeyDown events, but aren't bindable and should be discarded.
local unbindableKeys = {
	LALT = 1,
	LCTRL = 1,
	LSHIFT = 1,
	RALT = 1,
	RCTRL = 1,
	RSHIFT = 1,
	UNKNOWN = 1,
}


-- rounds off to N significant decimal places.
-- e.g. 5.975, 2 -> 5.98
local function roundOff(val, nPlaces)
	coef = 1
	if nPlaces then
		coef = 10^nPlaces
	end
	return floor((val*coef) + 0.5) / coef
end


-- Create the main QT button, and return it.
-- This is the button that holds the macrotext, and assigns a keybind to it when visible.
local function createMainButton()
	f = CreateFrame("Button", "QTMacrosMainButton", nil,
		"SecureHandlerShowHideTemplate,ActionButtonTemplate,SecureActionButtonTemplate")
	f.icon = _G[f:GetName().."Icon"]
	f.hotkey = _G[f:GetName().."HotKey"]
	f:SetFrameStrata("HIGH")
	f:SetMovable(true)
	f:SetClampedToScreen(true)
	f:Hide()

	if not db.buttonAnchor then
		db.buttonAnchor = { "CENTER", -100, -30 }
	end
	f:SetPoint(unpack(db.buttonAnchor))
	f:SetAttribute("type", "macro")
	if db.macroText then
		f:SetAttribute("macrotext", db.macroText)
	end
	f:SetAttribute("_onshow", [[
		local keybind = self:GetAttribute("keybind")
		if keybind then
			self:SetBindingClick(true, keybind, self:GetName(), "LeftButton")
		end
	]])
	f:SetAttribute("_onHide", [[
		self:ClearBindings()
	]])
	f:SetNormalTexture(nil)

	-- f:RegisterForClicks("LeftButtonDown", "RightButtonDown")
	f:RegisterForClicks("LeftButtonUp", "RightButtonUp")
	f:RegisterForDrag("LeftButton")
	f.icon:SetTexture([[Interface\Addons\QuickTargetMacros\images\QT_icon.tga]])
	
	SecureHandlerWrapScript(f, "OnClick", f, [[
		if button == "RightButton" then
			self:Hide()
			return false
		end
	]])
	
	function f:ShowTooltip()
		if not InCombatLockdown() then
			GameTooltip:SetOwner(self, "ANCHOR_RIGHT");
			GameTooltip:SetText("Quick Target Macros")
			GameTooltip:AddLine(self:GetAttribute("macrotext") or "")
			GameTooltip:AddLine(" ")
			GameTooltip:AddDoubleLine(L["Click"], L["Target"])
			GameTooltip:AddDoubleLine(L["Right-Click"], L["Dismiss"])
			GameTooltip:AddDoubleLine(L["Alt-Click"], L["Move"])
			GameTooltip:Show();
		end
	end
	
	f:SetScript("OnEnter", function(self)
		self:ShowTooltip()
	end)
	
	f:SetScript("OnLeave", function(self)
		GameTooltip:Hide()
	end)
	
	f:SetScript("OnDragStart", function(self)
		if IsAltKeyDown() then
			GameTooltip:Hide()
			self:StartMoving()
		end
	end)
	
	f:SetScript("OnDragStop", function(self)
		self:StopMovingOrSizing()
		self:SetUserPlaced(false)
		if self:GetNumPoints() > 0 then
			local point,_,_,x,y = self:GetPoint(1)
			db.buttonAnchor[1] = point
			db.buttonAnchor[2] = roundOff(x, 2)
			db.buttonAnchor[3] = roundOff(y, 2)
		end
		self:ShowTooltip()
	end)

	return f
end


-- You can call this multiple times. It only does init the first time.
function ShowKeybindFrame(...) -- declared local
	local f = CreateFrame("Frame")
	f:SetFrameStrata("DIALOG")
	f:EnableKeyboard(true)
	f:SetClampedToScreen(true)
	f:Hide()
	f:SetPoint("CENTER", 0, UIParent:GetHeight() / 7)
	f:SetBackdrop({
		bgFile=[[Interface\Tooltips\UI-Tooltip-Background]],
		edgeFile=[[Interface\Tooltips\UI-Tooltip-Border]], tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	f:SetBackdropColor(0, 0, 0)

	f.text = f:CreateFontString(nil, "ARTWORK", "GameFontNormal")
	f.text:SetPoint("TOP", 0, -20)
	f.text:SetText(L[ [=[
Press a key to bind the /target macro.

Or press ESC to cancel.]=] ])

	f.text2 = f:CreateFontString(nil, "ARTWORK", "GameFontNormalSmall")
	f.text2:SetPoint("BOTTOM", 0, 20)
	f.text2:SetText(L[ [=[
(This is a temporary "override" that only lasts as long as you're
using the macro. It won't remove your previous binding.)]=] ])

	f:SetScript("OnShow", function(self)
		local width = max(self.text:GetWidth(), self.text2:GetWidth()) + 40
		local height = self.text:GetHeight() + self.text2:GetHeight() + 60
		self:SetWidth(width)
		self:SetHeight(height)
		
		self:RegisterEvent("PLAYER_REGEN_DISABLED") -- so we can hide in combat
	end)
	
	f:SetScript("OnHide", function(self)
		self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	end)
	
	f:SetScript("OnEvent", function(self, event, ...)
		if event == "PLAYER_REGEN_DISABLED" then
			QTMacros:Print(L["Combat! Aborting keybind selection."])
			self:Hide() -- hide in combat
		end
	end)

	f:SetScript("OnKeyDown", function(self, key)
		if unbindableKeys[key] then
			return
		end
		if key == "ESCAPE" then
			self:Hide()
			return
		end
		if GetBindingFromClick(key) == "SCREENSHOT" then
			TakeScreenshot()
			return
		end
		if IsShiftKeyDown() then
			key = "SHIFT-"..key
		end
		if IsControlKeyDown() then
			key = "CTRL-"..key
		end
		if IsAltKeyDown() then
			key = "ALT-"..key
		end

		QTMacros:SetKeybind(key)
		self:Hide()
		if self.showButtonAfter then
			QTMacros:ActivateButton()
			self.showButtonAfter = nil
		end
	end)
	
	keybindFrame = f
	ShowKeybindFrame = function(showButtonAfter)
		keybindFrame.showButtonAfter = (showButtonAfter and 1 or nil)
		keybindFrame:Show()
	end
	return ShowKeybindFrame(...)
end


function QTMacros:SetKeybind(key)
	mainButton:SetAttribute("keybind", key)
	db.keybind = key
	if key then
		mainButton.hotkey:SetText(GetBindingText(key, "KEY_", true))
	else
		mainButton.hotkey:SetText(nil)
	end
end


function QTMacros:ActivateButton()
	if mainButton:GetAttribute("keybind") then
		mainButton:Show()
	else
		ShowKeybindFrame(true)
	end
end


function QTMacros:SlashQT(input)
	local targets = {strsplit(",", input)}
	for i = 1,#targets do
		targets[i] = strtrim(targets[i])
	end
	macroText = BuildMacro(db.cmdTemplate, targets)
	mainButton:SetAttribute("macrotext", macroText)
	db.macroText = macroText
	self:ActivateButton()
end


function QTMacros:OnChatCommand(input)
	LibStub("AceConfigCmd-3.0"):HandleCommand("qtmacros", "QTMacros", input)
end


function QTMacros:PLAYER_LOGOUT()
	-- db.isShown
	db.isShown = mainButton:IsShown()
end


function QTMacros:UpgradeDatabase()
	if db.db_version == nil then
		-- The first real version is 1. Old experimental versions had no db_version; we treat them as 0.
		db.db_version = 0
	end
	if db.db_version < 1 then
		-- This was for saving an actionID to drag a macro onto the bars in its place. Version 1 no longer uses it.
		db.savedActionID = nil
	end
	db.db_version = QTM_DB_VERSION
end


function QTMacros:OnInitialize()
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("QTMacros", giveOptions)
	local optFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("QTMacros", "QuickTargetMacros")

	-- LibStub("AceConfig-3.0"):RegisterOptionsTable("QuickTargetMacros", configTable, {"/quicktargetmacros", "/qtmacros", "/qtm"})
	self:RegisterChatCommand("quicktargetmacros", "OnChatCommand", true)
	self:RegisterChatCommand("qtmacros", "OnChatCommand", true)
	self:RegisterChatCommand("qtm", "OnChatCommand", true)

	self:RegisterChatCommand("qt", "SlashQT")
	self:RegisterChatCommand("qtmobname", "SlashQT")
	self:RegisterChatCommand("qtname", "SlashQT")

	QTMacrosPerCharDB = QTMacrosPerCharDB or {}
	db = QTMacrosPerCharDB
	if db.db_version ~= QTM_DB_VERSION then
		self:UpgradeDatabase()
	end
	if not db.cmdTemplate then
		db.cmdTemplate = {
			line = "/tar $s",
		}
	end

	mainButton = createMainButton()
	createMainButton = nil

	if db.keybind then
		self:SetKeybind(db.keybind)
	end
	if db.isShown then
		mainButton:Show()
	end
end


function QTMacros:OnEnable()
	self:RegisterEvent("PLAYER_LOGOUT")
end


function QTMacros:OnDisable()
end
